@@ -1,28 +1,27 @@
 /*****************************************************************************\
- *  $Id: fillfile.c 77 2006-02-15 01:00:42Z garlick $
- *****************************************************************************
- *  Copyright (C) 2001-2008 The Regents of the University of California.
+ *  Copyright (C) 2001-2007 The Regents of the University of California.
+ *  Copyright (C) 2007-2014 Lawrence Livermore National Security, LLC.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
- *  Written by Jim Garlick <garlick@llnl.gov>.
- *  UCRL-CODE-2003-006.
- *  
+ *  Written by Jim Garlick <garlick@llnl.gov> UCRL-CODE-2003-006
+ *
  *  This file is part of Scrub, a program for erasing disks.
- *  For details, see https://code.google.com/p/diskscrub/
- *  
- *  Scrub is free software; you can redistribute it and/or modify it under
- *  the terms of the GNU General Public License as published by the Free
- *  Software Foundation; either version 2 of the License, or (at your option)
+ *  For details, see http://code.google.com/p/diskscrub.
+ *
+ *  This program is free software; you can redistribute it and/or modify it
+ *  under the terms of the GNU General Public License as published by the Free
+ *  Software Foundation; either version 2 of the license, or (at your option)
  *  any later version.
- *  
- *  Scrub is distributed in the hope that it will be useful, but WITHOUT ANY
- *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
- *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
- *  details.
- *  
+ *
+ *  This program is distributed in the hope that it will be useful, but
+ *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+ *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+ *  for more details.
+ *
  *  You should have received a copy of the GNU General Public License along
- *  with Scrub; if not, write to the Free Software Foundation, Inc.,
- *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
-\*****************************************************************************/
+ *  with this program; if not, write to the Free Software Foundation,
+ *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
+ *  See also: http://www.gnu.org/licenses
+ *****************************************************************************/
 
 #if HAVE_CONFIG_H
 #include "config.h"
@@ -43,6 +42,8 @@
 #include "util.h"
 #include "fillfile.h"
 
+static int no_threads = 0;
+
 struct memstruct {
     refill_t refill;
     unsigned char *buf;
@@ -50,7 +51,6 @@ struct memstruct {
 #if WITH_PTHREADS
     pthread_t thd;
     int err;
-    bool running;
 #endif
 };
 
@@ -62,13 +62,6 @@ extern char *prog;
 # define MY_O_DIRECT 0
 #endif
 
-static void
-oom(void)
-{
-    fprintf (stderr, "%s: out of memory\n", prog);
-    exit (1);
-}
-
 static void *
 refill_thread(void *arg)
 {
@@ -78,62 +71,77 @@ refill_thread(void *arg)
     return mp;
 }
 
-static void
+static int
 refill_memcpy(struct memstruct *mp, unsigned char *mem, int memsize,
-              int filesize, int written)
+              off_t filesize, off_t written)
 {
 #if WITH_PTHREADS
-    if ((mp->err = pthread_join(mp->thd, NULL))) {
-        fprintf(stderr, "%s: pthread_join: %s\n", prog, strerror(mp->err));
-        exit(1);
+    if (no_threads) {
+        mp->size = memsize;
+        refill_thread (mp);
+    } else {
+        if ((mp->err = pthread_join(mp->thd, NULL))) {
+            errno = mp->err;
+            goto error;
+        }
+        assert (memsize == mp->size);
     }
-    mp->running = false;
 #else
+    mp->size = memsize;
     refill_thread (mp);
 #endif
-    assert (memsize == mp->size);
     memcpy(mem, mp->buf, memsize);
 #if WITH_PTHREADS
-    written += memsize;
-    if (filesize - written > 0) {
-        if (mp->size > filesize - written)
-            mp->size = filesize - written;
-        mp->running = true;
-        if ((mp->err = pthread_create(&mp->thd, NULL, refill_thread, mp))) {
-            fprintf(stderr, "%s: pthread_create: %s\n", prog,
-                    strerror(mp->err));
-            exit(1);
+    if (!no_threads) {
+        written += memsize;
+        if (filesize - written > 0) {
+            if (mp->size > filesize - written)
+                mp->size = filesize - written;
+            if ((mp->err = pthread_create(&mp->thd, NULL, refill_thread, mp))) {
+                errno = mp->err;
+                goto error;
+            }
         }
     }
 #endif
+    return 0;
+error:
+    return -1;
 }
 
-static void
+static int
 refill_init(struct memstruct **mpp, refill_t refill, int memsize)
 {
     struct memstruct *mp = NULL;
 
     if (!(mp = malloc(sizeof(struct memstruct))))
-        oom();
+        goto nomem;
     if (!(mp->buf = malloc(memsize)))
-        oom();
+        goto nomem;
     mp->size = memsize;
     mp->refill = refill;
 #if WITH_PTHREADS
-    mp->running = true;
-    if ((mp->err = pthread_create(&mp->thd, NULL, refill_thread, mp))) {
-        fprintf(stderr, "%s: pthread_create: %s\n", prog, strerror(mp->err));
-        exit(1);
+    if (!no_threads) {
+        if ((mp->err = pthread_create(&mp->thd, NULL, refill_thread, mp))) {
+            errno = mp->err;
+            goto error;
+        }
     }
 #endif
     *mpp = mp;
+    return 0;
+nomem:
+    errno = ENOMEM;
+error:
+    return -1;
 }
 
 static void
 refill_fini(struct memstruct *mp)
 {
 #if WITH_PTHREADS
-    assert(mp->running == false);
+    if (!no_threads)
+        (void)pthread_join(mp->thd, NULL);
 #endif
     free (mp->buf);
     free (mp);
@@ -152,7 +160,7 @@ fillfile(char *path, off_t filesize, unsigned char *mem, int memsize,
          progress_t progress, void *arg, refill_t refill, 
          bool sparse, bool creat)
 {
-    int fd;
+    int fd = -1;
     off_t n;
     off_t written = 0LL;
     int openflags = O_WRONLY;
@@ -168,62 +176,56 @@ fillfile(char *path, off_t filesize, unsigned char *mem, int memsize,
         openflags &= ~MY_O_DIRECT;
         fd = open(path, openflags, 0644);
     }
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
+    if (fd < 0)
+        goto error;
     do {
         if (written + memsize > filesize)
             memsize = filesize - written;
         if (refill && !sparse) {
             if (!mp)
-                refill_init(&mp, refill, memsize);
-            refill_memcpy(mp, mem, memsize, filesize, written);
+                if (refill_init(&mp, refill, memsize) < 0)
+                    goto error;
+            if (refill_memcpy(mp, mem, memsize, filesize, written) < 0)
+                goto error;
         }
         if (sparse && !(written == 0) && !(written + memsize == filesize)) {
-            if (lseek(fd, memsize, SEEK_CUR) < 0) {
-                fprintf(stderr, "%s: lseek %s: %s\n", prog, path, 
-                        strerror(errno));
-                exit(1);
-            }
+            if (lseek(fd, memsize, SEEK_CUR) < 0)
+                goto error;
             written += memsize;
         } else {
             n = write_all(fd, mem, memsize);
             if (creat && n < 0 && errno == ENOSPC)
                 break;
             if (n == 0) {
-                fprintf(stderr, "%s: write %s: %s\n", prog, path,
-                    "returned zero - attempt to write past end of device?");
-                exit(1);
-            }
-            if (n < 0) {
-                fprintf(stderr, "%s: write %s: %s\n", prog, path, 
-                        strerror(errno));
-                exit(1);
-            }
+                errno = EINVAL; /* write past end of device? */
+                goto error;
+            } else if (n < 0)
+                goto error;
             written += n;
         }
         if (progress)
             progress(arg, (double)written/filesize);
     } while (written < filesize);
     if (fsync(fd) < 0) {
-        fprintf(stderr, "%s: fsync %s: %s\n", prog, path, strerror(errno));
-        exit(1);
+        if (errno != EINVAL)
+            goto error;
+        errno = 0;
     }
 #if defined(HAVE_POSIX_FADVISE) && defined(POSIX_FADV_DONTNEED)
     /* Try to fool the kernel into dropping any device cache */
-    if (posix_fadvise(fd, 0, filesize, POSIX_FADV_DONTNEED)) {
-        fprintf(stderr, "%s: posix_fadvise %s: %s\n",
-                prog, path, strerror(errno));
-    }
+    (void)posix_fadvise(fd, 0, filesize, POSIX_FADV_DONTNEED);
 #endif
-    if (close(fd) < 0) {
-        fprintf(stderr, "%s: close %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
-    if (refill && !sparse)
+    if (close(fd) < 0)
+        goto error;
+    if (mp)
         refill_fini(mp);
     return written;
+error:
+    if (mp)
+        refill_fini(mp);
+    if (fd != -1)
+        (void)close(fd);
+    return (off_t)-1;
 }
 
 /* Verify that file was filled with 'mem' patterns.
@@ -232,17 +234,14 @@ off_t
 checkfile(char *path, off_t filesize, unsigned char *mem, int memsize,
           progress_t progress, void *arg, bool sparse)
 {
-    int fd;
+    int fd = -1;
     off_t n;
     off_t verified = 0LL;
-    unsigned char *buf;
+    unsigned char *buf = NULL;
     int openflags = O_RDONLY;
 
-    buf = alloc_buffer(memsize);
-    if (!buf) {
-        fprintf(stderr, "out of memory\n");
-        exit(1);
-    }
+    if (!(buf = alloc_buffer(memsize)))
+        goto nomem;
     if (filetype(path) != FILE_CHAR)
         openflags |= MY_O_DIRECT;
     fd = open(path, openflags);
@@ -251,45 +250,49 @@ checkfile(char *path, off_t filesize, unsigned char *mem, int memsize,
         openflags &= ~MY_O_DIRECT;
         fd = open(path, openflags);
     }
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
+    if (fd < 0)
+        goto error;
     do {
         if (verified + memsize > filesize)
             memsize = filesize - verified;
         if (sparse && !(verified == 0) && !(verified + memsize == filesize)) {
-            if (lseek(fd, memsize, SEEK_CUR) < 0) {
-                fprintf(stderr, "%s: lseek %s: %s\n", prog, path, 
-                        strerror(errno));
-                exit(1);
-            }
+            if (lseek(fd, memsize, SEEK_CUR) < 0)
+                goto error;
             verified += memsize;
         } else {
             n = read_all(fd, buf, memsize);
-            if (n < 0) {
-                fprintf(stderr, "%s: read %s: %s", prog, path, strerror(errno));
-                exit(1);
-            }
+            if (n < 0)
+                goto error;
             if (n == 0) {
-                fprintf(stderr, "%s: premature EOF on %s\n", prog, path);
-                exit(1);
+                errno = EINVAL; /* early EOF */
+                goto error;
             }
             if (memcmp(mem, buf, memsize) != 0) {
-                fprintf(stderr, "%s: verification failure on %s\n", prog, path);
-                exit(1);
+                break; /* return < filesize means verification failure */
             }
             verified += n;
         }
         if (progress)
             progress(arg, (double)verified/filesize);
     } while (verified < filesize);
-    if (close(fd) < 0) {
-        fprintf(stderr, "%s: close %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
+    if (close(fd) < 0)
+        goto error;
     free(buf);
     return verified;
+nomem:
+    errno = ENOMEM;
+error:
+    if (buf)
+        free (buf);
+    if (fd != -1)
+        (void)close(fd);
+    return (off_t)-1;
+}
+
+void
+disable_threads(void)
+{
+    no_threads = 1;
 }
 
 /*
