@@ -1,28 +1,27 @@
 /*****************************************************************************\
- *  $Id: getsize.c 76 2006-02-15 00:49:19Z garlick $
- *****************************************************************************
- *  Copyright (C) 2001-2008 The Regents of the University of California.
+ *  Copyright (C) 2001-2007 The Regents of the University of California.
+ *  Copyright (C) 2007-2014 Lawrence Livermore National Security, LLC.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
- *  Written by Jim Garlick <garlick@llnl.gov>.
- *  UCRL-CODE-2003-006.
- *  
+ *  Written by Jim Garlick <garlick@llnl.gov> UCRL-CODE-2003-006
+ *
  *  This file is part of Scrub, a program for erasing disks.
- *  For details, see https://code.google.com/p/diskscrub/
- *  
- *  Scrub is free software; you can redistribute it and/or modify it under
- *  the terms of the GNU General Public License as published by the Free
- *  Software Foundation; either version 2 of the License, or (at your option)
+ *  For details, see http://code.google.com/p/diskscrub.
+ *
+ *  This program is free software; you can redistribute it and/or modify it
+ *  under the terms of the GNU General Public License as published by the Free
+ *  Software Foundation; either version 2 of the license, or (at your option)
  *  any later version.
- *  
- *  Scrub is distributed in the hope that it will be useful, but WITHOUT ANY
- *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
- *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
- *  details.
- *  
+ *
+ *  This program is distributed in the hope that it will be useful, but
+ *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+ *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+ *  for more details.
+ *
  *  You should have received a copy of the GNU General Public License along
- *  with Scrub; if not, write to the Free Software Foundation, Inc.,
- *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
-\*****************************************************************************/
+ *  with this program; if not, write to the Free Software Foundation,
+ *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
+ *  See also: http://www.gnu.org/licenses
+ *****************************************************************************/
 
 #if HAVE_CONFIG_H
 #include "config.h"
@@ -49,19 +48,16 @@ extern char *prog;
 #include <sys/utsname.h>
 typedef unsigned long long u64; /* for BLKGETSIZE64 (slackware) */
 
-off_t 
-getsize(char *path)
+int
+getsize(char *path, off_t *sizep)
 {
     struct utsname ut;
     unsigned long long numbytes;
     int valid_blkgetsize64 = 1;
-    int fd;
+    int fd = -1;
 
-    fd = open(path, O_RDONLY);
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
+    if ((fd = open(path, O_RDONLY)) < 0)
+        goto error;
 
     /* Ref: e2fsprogs-1.39 - apparently BLKGETSIZE64 doesn't work pre 2.6 */
     if ((uname(&ut) == 0) &&
@@ -69,25 +65,24 @@ getsize(char *path)
              (ut.release[2] < '6') && (ut.release[3] == '.')))
                 valid_blkgetsize64 = 0;
     if (valid_blkgetsize64) {
-        if (ioctl(fd, BLKGETSIZE64, &numbytes) < 0) {
-            fprintf(stderr, "%s: ioctl BLKGETSIZE64 %s: %s\n", prog, path,
-                    strerror(errno));
-            exit(1);
-        }
+        if (ioctl(fd, BLKGETSIZE64, &numbytes) < 0)
+            goto error;
     } else {
         unsigned long numblocks;
 
-        if (ioctl(fd, BLKGETSIZE, &numblocks) < 0) {
-            fprintf(stderr, "%s: ioctl BLKGETSIZE %s: %s\n", prog, path,
-                    strerror(errno));
-            exit(1);
-        }
+        if (ioctl(fd, BLKGETSIZE, &numblocks) < 0)
+            goto error;
         numbytes = (off_t)numblocks*512; /* 2TB limit here */
     }
 
-    (void)close(fd);
-
-    return (off_t)numbytes;
+    if (close(fd) < 0)
+        goto error;
+    *sizep = (off_t)numbytes;
+    return 0;
+error:
+    if (fd != -1)
+        (void)close(fd);
+    return -1;
 }
 
 #elif defined(__FreeBSD__)
@@ -95,25 +90,24 @@ getsize(char *path)
 #include <sys/ioctl.h>
 #include <sys/disk.h>
 
-off_t 
-getsize(char *path)
+int
+getsize(char *path, off_t *sizep)
 {
     off_t numbytes;
-    int fd;
+    int fd = -1;
 
-    fd = open(path, O_RDONLY);
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
-    if (ioctl(fd, DIOCGMEDIASIZE, &numbytes) < 0) {
-        fprintf(stderr, "%s: ioctl DIOCGMEDIASIZE %s: %s\n", prog, path,
-                strerror(errno));
-        exit(1);
-    }
-    (void)close(fd);
-
-    return numbytes;
+    if ((fd = open(path, O_RDONLY)) < 0)
+        goto error;
+    if (ioctl(fd, DIOCGMEDIASIZE, &numbytes) < 0)
+        goto error;
+    if (close(fd) < 0)
+        goto error;
+    *sizep = numbytes;
+    return 0;
+error:
+    if (fd != -1)
+        (void)close(fd);
+    return -1;
 }
 
 #elif defined(sun)
@@ -122,25 +116,24 @@ getsize(char *path)
 #include <sys/dkio.h>
 #include <sys/vtoc.h>
 
-off_t 
-getsize(char *path)
+int
+getsize(char *path, off_t *sizep)
 {
     struct dk_minfo dkmp;
-    int fd;
+    int fd = -1;
 
-    fd = open(path, O_RDONLY);
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
-    if (ioctl(fd, DKIOCGMEDIAINFO, &dkmp) < 0) {
-        fprintf(stderr, "%s: ioctl DKIOCGMEDIAINFO %s: %s\n", prog, path,
-                strerror(errno));
-        exit(1);
-    }
-    (void)close(fd);
-
-    return (off_t)dkmp.dki_capacity * dkmp.dki_lbsize;
+    if ((fd = open(path, O_RDONLY)) < 0)
+        goto error;
+    if (ioctl(fd, DKIOCGMEDIAINFO, &dkmp) < 0)
+        goto error;
+    if (close(fd) < 0)
+        goto error;
+    *sizep = (off_t)dkmp.dki_capacity * dkmp.dki_lbsize;
+    return 0;
+error:
+    if (fd != -1)
+        (void)close(fd);
+    return -1;
 }
 
 #elif defined(__APPLE__)
@@ -149,31 +142,27 @@ getsize(char *path)
 #include <sys/ioctl.h>
 #include <sys/disk.h>
 
-off_t 
-getsize(char *path)
+int
+getsize(char *path, off_t *sizep)
 {
     uint32_t blocksize;
     uint64_t blockcount;
-    int fd;
+    int fd = -1;
 
-    fd = open(path, O_RDONLY);
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
-    if (ioctl(fd, DKIOCGETBLOCKSIZE, &blocksize) < 0) {
-        fprintf(stderr, "%s: ioctl DKIOCGETBLOCKSIZE %s: %s\n", prog, path,
-                strerror(errno));
-        exit(1);
-    }
-    if (ioctl(fd, DKIOCGETBLOCKCOUNT, &blockcount) < 0) {
-        fprintf(stderr, "%s: ioctl DKIOGETBLOCKCOUNT %s: %s\n", prog, path,
-                strerror(errno));
-        exit(1);
-    }
-    (void)close(fd);
-
-    return (off_t)blockcount * blocksize;
+    if ((fd = open(path, O_RDONLY)) < 0)
+        goto error;
+    if (ioctl(fd, DKIOCGETBLOCKSIZE, &blocksize) < 0)
+        goto error;
+    if (ioctl(fd, DKIOCGETBLOCKCOUNT, &blockcount) < 0)
+        goto error;
+    if (close(fd) < 0)
+        goto error;
+    *sizep = (off_t)blockcount * blocksize;
+    return 0;
+error:
+    if (fd != -1)
+        (void)close(fd);
+    return -1;
 }
 #elif defined(_AIX)
 /* scrub-1.7 tested AIX 5.1 and 5.3 */
@@ -181,62 +170,65 @@ getsize(char *path)
 #include <sys/ioctl.h>
 #include <sys/devinfo.h>
 
-off_t 
-getsize(char *path)
+int
+getsize(char *path, off_t *sizep)
 {
-    int fd;
+    int fd = -1;
     struct devinfo devinfo;
     off_t size;
 
-    fd = open(path, O_RDONLY);
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
-    if (ioctl(fd, IOCINFO, &devinfo) == -1) {
-        fprintf(stderr, "%s: ioctl IOCINFO %s: %s\n", prog, path, 
-                strerror(errno));
-        exit(1);
-    }
+    if ((fd = open(path, O_RDONLY)) < 0)
+        goto error;
+    if (ioctl(fd, IOCINFO, &devinfo) == -1)
+        goto error;
     switch (devinfo.devtype) {
         case DD_DISK:   /* disk */
             size = (off_t)devinfo.un.dk.segment_size * devinfo.un.dk.segment_count;
             break;
         case DD_SCDISK: /* scsi disk */
+#ifdef DF_LGDSK
+            /* T.P.Starrin@NASA.gov for AIX 6.1 */
+            size = (off_t)((unsigned)(devinfo.un.scdk64.hi_numblks << 32) | ((unsigned)devinfo.un.scdk64.lo_numblks)) * devinfo.un.scdk64.blksize;
+#else
             size = (off_t)devinfo.un.scdk.blksize * devinfo.un.scdk.numblks;
+#endif
             break;
         default:        /* unknown */
             size = 0;
             break;
     }
-
-    (void)close(fd);
-    return size;
+    if (close(fd) < 0)
+        goto error;
+    *sizep = size;
+    return 0;
+error:
+    if (fd != -1)
+        (void)close(fd);
+    return -1;
 }
 #elif defined (__hpux)
 
 #include <stropts.h>
 #include <sys/scsi.h>
 
-off_t
-getsize(char *path)
+int
+getsize(char *path, off_t *sizep)
 {
-    int fd;
+    int fd = -1;
     struct capacity cap;
 
-    fd = open(path, O_RDONLY);
-    if (fd < 0) {
-        fprintf(stderr, "%s: open %s: %s\n", prog, path, strerror(errno));
-        exit(1);
-    }
-    if (ioctl(fd, SIOC_CAPACITY, &cap) == -1) {
-        fprintf(stderr, "%s: ioctl SIOC_CAPACITY %s: %s\n", prog, path,
-                strerror(errno));
-        exit(1);
-    }
-
-    (void)close(fd);
-    return (off_t)cap.lba * cap.blksz;
+    if ((fd = open(path, O_RDONLY)) < 0)
+        goto error;
+    if (ioctl(fd, SIOC_CAPACITY, &cap) == -1)
+        goto error;
+    if (close(fd) < 0)
+        goto error;
+    *sizep = (off_t)cap.lba * cap.blksz;
+    return 0;
+error:
+    if (fd != -1)
+        (void)close(fd);
+    return -1;
 }
 
 #else
@@ -245,7 +237,8 @@ getsize(char *path)
 off_t 
 getsize(char *path)
 {
-    return 0;
+    errno = ENOSYS;
+    return -1;
 }
 #endif
 
@@ -263,15 +256,15 @@ size2str(char *str, int len, off_t size)
 
     if (eb >= 1) {
         num = eb; unit = "EB";
-    } else if (pb >= 1) {
+    } else if (pb >= 10) {
         num = pb; unit = "PB";
-    } else if (tb >= 1) {
+    } else if (tb >= 10) {
         num = tb; unit = "TB";
-    } else if (gb >= 1) {
+    } else if (gb >= 10) {
         num = gb; unit = "GB";
-    } else if (mb >= 1) {
+    } else if (mb >= 10) {
         num = mb; unit = "MB";
-    } else if (kb >= 1) {
+    } else if (kb >= 10) {
         num = kb; unit = "KB";
     } 
    
