@@ -1,28 +1,27 @@
 /*****************************************************************************\
- *  $Id: scrub.c 78 2006-02-15 01:05:03Z garlick $
- *****************************************************************************
- *  Copyright (C) 2001-2008 The Regents of the University of California.
+ *  Copyright (C) 2001-2007 The Regents of the University of California.
+ *  Copyright (C) 2007-2014 Lawrence Livermore National Security, LLC.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
- *  Written by Jim Garlick <garlick@llnl.gov>.
- *  UCRL-CODE-2003-006.
- *  
+ *  Written by Jim Garlick <garlick@llnl.gov> UCRL-CODE-2003-006
+ *
  *  This file is part of Scrub, a program for erasing disks.
- *  For details, see https://code.google.com/p/diskscrub/
- *  
- *  Scrub is free software; you can redistribute it and/or modify it under
- *  the terms of the GNU General Public License as published by the Free
- *  Software Foundation; either version 2 of the License, or (at your option)
+ *  For details, see http://code.google.com/p/diskscrub.
+ *
+ *  This program is free software; you can redistribute it and/or modify it
+ *  under the terms of the GNU General Public License as published by the Free
+ *  Software Foundation; either version 2 of the license, or (at your option)
  *  any later version.
- *  
- *  Scrub is distributed in the hope that it will be useful, but WITHOUT ANY
- *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
- *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
- *  details.
- *  
+ *
+ *  This program is distributed in the hope that it will be useful, but
+ *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+ *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+ *  for more details.
+ *
  *  You should have received a copy of the GNU General Public License along
- *  with Scrub; if not, write to the Free Software Foundation, Inc.,
- *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
-\*****************************************************************************/
+ *  with this program; if not, write to the Free Software Foundation,
+ *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
+ *  See also: http://www.gnu.org/licenses
+ *****************************************************************************/
 
 /* Scrub a raw disk or plain file.
  */
@@ -57,22 +56,33 @@
 
 #define BUFSIZE (4*1024*1024) /* default blocksize */
 
-static void       usage(void);
-static bool       scrub(char *path, off_t size, const sequence_t *seq,
-                      int bufsize, bool Sopt, bool sparse, bool enospc);
-static void       scrub_free(char *path, off_t size, const sequence_t *seq,
-                      int bufsize, bool Sopt);
-static void       scrub_dirent(char *path, char *newpath);
-static void       scrub_file(char *path, off_t size, const sequence_t *seq,
-                      int bufsize, bool Sopt, bool sparse);
-#if __APPLE__
-static void       scrub_resfork(char *path, const sequence_t *seq,
-                      int bufsize);
-#endif
-static void       scrub_disk(char *path, off_t size, const sequence_t *seq,
-                      int bufsize, bool Sopt, bool sparse);
+struct opt_struct {
+    const sequence_t *seq;
+    int blocksize;
+    off_t devsize;
+    char *dirent;
+    bool force;
+    bool nosig;
+    bool remove;
+    bool sparse;
+    bool nofollow;
+    bool nohwrand;
+    bool nothreads;
+};
 
-#define OPTIONS "p:D:Xb:s:fSrvTLRh"
+static bool       scrub(char *path, off_t size, const sequence_t *seq,
+                      int bufsize, bool nosig, bool sparse, bool enospc);
+static void       scrub_free(char *path, const struct opt_struct *opt);
+static void       scrub_dirent(char *path, const struct opt_struct *opt);
+static void       scrub_file(char *path, const struct opt_struct *opt);
+#if __APPLE__
+static void       scrub_resfork(char *path, const struct opt_struct *opt);
+#endif
+static void       scrub_disk(char *path, const struct opt_struct *opt);
+static int        scrub_object(char *path, const struct opt_struct *opt,
+                               bool noexec, bool dryrun);
+
+#define OPTIONS "p:D:Xb:s:fSrvTLRthn"
 #if HAVE_GETOPT_LONG
 #define GETOPT(ac,av,opt,lopt) getopt_long(ac,av,opt,lopt,NULL)
 static struct option longopts[] = {
@@ -88,6 +98,8 @@ static struct option longopts[] = {
     {"test-sparse",      no_argument,        0, 'T'},
     {"no-link",          no_argument,        0, 'L'},
     {"no-hwrand",        no_argument,        0, 'R'},
+    {"no-threads",       no_argument,        0, 't'},
+    {"dry-run",          no_argument,        0, 'n'},
     {"help",             no_argument,        0, 'h'},
     {0, 0, 0, 0},
 };
@@ -97,195 +109,11 @@ static struct option longopts[] = {
 
 char *prog;
 
-int 
-main(int argc, char *argv[])
-{
-    const sequence_t *seq = NULL;
-    bool Xopt = false;
-    int bopt = BUFSIZE;
-    off_t sopt = 0;
-    char *Dopt = NULL;
-    char *filename = NULL;
-    bool fopt = false;
-    bool Sopt = false;
-    bool ropt = false;
-    bool Topt = false;
-    bool Lopt = false;
-    bool Ropt = false;
-    extern int optind;
-    extern char *optarg;
-    int c;
-
-    assert(sizeof(off_t) == 8);
-
-    /* Handle arguments.
-     */
-    prog = basename(argv[0]);
-    while ((c = GETOPT(argc, argv, OPTIONS, longopts)) != EOF) {
-        switch (c) {
-        case 'v':   /* --version */
-            printf("scrub version %s\n", VERSION);
-            exit(0);
-        case 'p':   /* --pattern */
-            seq = seq_lookup(optarg);
-            if (!seq) {
-                fprintf(stderr, "%s: no such pattern sequence\n", prog);
-                exit(1);
-            }
-            break;
-        case 'X':   /* --freespace */
-            Xopt = true;
-            break;
-        case 'D':   /* --dirent */
-            Dopt = optarg;
-            break;
-        case 'r':   /* --remove */
-            ropt = true;
-            break;
-        case 'b':   /* --blocksize */
-            bopt = str2int(optarg);
-            if (bopt == 0) {
-                fprintf(stderr, "%s: error parsing blocksize string\n", prog);
-                exit(1);
-            }
-            break;
-        case 's':   /* --device-size */
-            sopt = str2size(optarg);
-            if (sopt == 0) {
-                fprintf(stderr, "%s: error parsing size string\n", prog);
-                exit(1);
-            }
-            break;
-        case 'f':   /* --force */
-            fopt = true;
-            break;
-        case 'S':   /* --no-signature */
-            Sopt = true;
-            break;
-        case 'T':   /* --test-sparse */
-            Topt = true;
-            break;
-        case 'L':   /* --no-link */
-            Lopt = true;
-            break;
-        case 'R':   /* --no-hwrand */
-            Ropt = true;
-            break;
-        case 'h':   /* --help */
-        default:
-            usage();
-        }
-    }
-    if (argc - optind != 1)
-        usage();
-    filename = argv[optind];
-
-    if (!seq)
-        seq = seq_lookup("nnsa");
-    assert(seq != NULL);
-    printf("%s: using %s patterns\n", prog, seq->desc);
-
-    if (Ropt)
-        disable_hwrand();
-
-    /* Handle -X specially.
-     */
-    if (Xopt) {
-        if (filetype(filename) != FILE_NOEXIST) {
-            fprintf(stderr, "%s: -X directory already exists\n", prog);
-            exit(1);
-        }
-        if (Dopt) {
-            fprintf(stderr, "%s: -D and -X cannot be used together\n", prog);
-            exit(1);
-        }
-        scrub_free(filename, sopt, seq, bopt, Sopt);
-        goto done;
-    } 
-
-    switch (filetype(filename)) {
-        case FILE_NOEXIST:
-            fprintf(stderr, "%s: %s does not exist\n", prog, filename);
-            exit(1);
-        case FILE_OTHER:
-            fprintf(stderr, "%s: %s is wrong type of file\n", prog, filename);
-            exit(1);
-        case FILE_BLOCK:
-        case FILE_CHAR:
-            if (Dopt) {
-                fprintf(stderr, "%s: cannot use -D with special file\n", prog);
-                exit(1);
-            }
-            if (ropt) {
-                fprintf(stderr, "%s: cannot use -r with special file\n", prog);
-                exit(1);
-            }
-            if (access(filename, R_OK|W_OK) < 0) {
-                fprintf(stderr, "%s: no rw access to %s\n", prog, filename);
-                exit(1);
-            }
-            if (checksig(filename) && !fopt) {
-                fprintf(stderr, "%s: %s already scrubbed? (-f to force)\n",
-                        prog, filename);
-                exit(1);
-            }
-            scrub_disk(filename, sopt, seq, bopt, Sopt, Topt);
-            break;
-        case FILE_LINK:
-            if (Lopt) {
-                if (ropt) {
-                    printf("%s: unlinking %s\n", prog, filename);
-                    if (unlink(filename) != 0) {
-                        fprintf(stderr, "%s: unlink %s: %s\n", prog, filename,
-                            strerror(errno));
-                        exit(1);
-                    }
-                }
-                break;
-            }
-        case FILE_REGULAR:
-            if (access(filename, R_OK|W_OK) < 0) {
-                fprintf(stderr, "%s: no rw access to %s\n", prog, filename);
-                exit(1);
-            }
-            if (checksig(filename) && !fopt) {
-                fprintf(stderr, "%s: %s already scrubbed? (-f to force)\n",
-                        prog, filename);
-                exit(1);
-            }
-            if (Dopt && *Dopt != '/' && *filename == '/') {
-                fprintf(stderr, "%s: %s should be a full path like %s\n",
-                        prog, Dopt, filename);
-                exit(1);
-            }
-            scrub_file(filename, sopt, seq, bopt, Sopt, Topt);
-#if __APPLE__
-            scrub_resfork(filename, seq, bopt);
-#endif
-            if (Dopt) {
-                scrub_dirent(filename, Dopt);
-                filename = Dopt; /* -r needs this below */
-            }
-            if (ropt) {
-                printf("%s: unlinking %s\n", prog, filename);
-                if (unlink(filename) != 0) {
-                    fprintf(stderr, "%s: unlink %s: %s\n", prog, filename,
-                            strerror(errno));
-                    exit(1);
-                }
-            }
-            break;
-    }
-done:
-    pattern_finalize ();
-    exit(0);
-}
-
 static void 
 usage(void)
 {
     fprintf(stderr,
-"Usage: %s [OPTIONS] file\n"
+"Usage: %s [OPTIONS] file [file...]\n"
 "  -v, --version           display scrub version and exit\n"
 "  -p, --pattern pat       select scrub pattern sequence\n"
 "  -b, --blocksize size    set I/O buffer size (default 4m)\n"
@@ -297,6 +125,8 @@ usage(void)
 "  -r, --remove            remove file after scrub\n"
 "  -L, --no-link           do not scrub link target\n"
 "  -R, --no-hwrand         do not use a hardware random number generator\n"
+"  -t, --no-threads        do not compute random data in a parallel thread\n"
+"  -n, --dry-run           verify file arguments, without writing\n"
 "  -h, --help              display this help message\n"
     , prog);
 
@@ -305,6 +135,291 @@ usage(void)
     exit(1);
 }
 
+int 
+main(int argc, char *argv[])
+{
+    struct opt_struct opt;
+    sequence_t *custom_seq = NULL;
+    bool Xopt = false;
+    bool nopt = false;
+    extern int optind;
+    extern char *optarg;
+    int c;
+
+    assert(sizeof(off_t) == 8);
+
+    memset (&opt, 0, sizeof (opt));
+    opt.blocksize = BUFSIZE;
+
+    /* Handle arguments.
+     */
+    prog = basename(argv[0]);
+    while ((c = GETOPT(argc, argv, OPTIONS, longopts)) != EOF) {
+        switch (c) {
+        case 'v':   /* --version */
+            printf("scrub version %s\n", VERSION);
+            exit(0);
+        case 'p':   /* --pattern */
+            if (opt.seq != NULL) {
+                fprintf(stderr, "%s: only one pattern can be selected\n", prog);
+                exit(1);
+            }
+            if (!strncmp (optarg, "custom=", 7) && strlen (optarg) > 7) {
+                if (!(custom_seq = seq_create ("custom", "Custom single-pass",
+                                               &optarg[7]))) {
+                    fprintf(stderr, "%s: custom sequence: %s\n", prog,
+                            strerror(errno));
+                    exit(1);
+                }
+                opt.seq = custom_seq;
+            } else {
+                opt.seq = seq_lookup(optarg);
+                if (!opt.seq) {
+                    fprintf(stderr, "%s: no such pattern sequence\n", prog);
+                    exit(1);
+                }
+            }
+            break;
+        case 'X':   /* --freespace */
+            Xopt = true;
+            break;
+        case 'D':   /* --dirent */
+            opt.dirent = optarg;
+            break;
+        case 'r':   /* --remove */
+            opt.remove = true;
+            break;
+        case 'b':   /* --blocksize */
+            opt.blocksize = str2int(optarg);
+            if (opt.blocksize == 0) {
+                fprintf(stderr, "%s: error parsing blocksize string\n", prog);
+                exit(1);
+            }
+            break;
+        case 's':   /* --device-size */
+            opt.devsize = str2size(optarg);
+            if (opt.devsize == 0) {
+                fprintf(stderr, "%s: error parsing size string\n", prog);
+                exit(1);
+            }
+            break;
+        case 'f':   /* --force */
+            opt.force = true;
+            break;
+        case 'S':   /* --no-signature */
+            opt.nosig = true;
+            break;
+        case 'T':   /* --test-sparse */
+            opt.sparse = true;
+            break;
+        case 'L':   /* --no-link */
+            opt.nofollow = true;
+            break;
+        case 'R':   /* --no-hwrand */
+            opt.nohwrand = true;
+            break;
+        case 't':   /* --no-threads */
+            opt.nothreads = true;
+            break;
+        case 'n':   /* --dry-run */
+            nopt = true;
+            break;
+        case 'h':   /* --help */
+        default:
+            usage();
+        }
+    }
+    if (argc == optind)
+        usage();
+    if (Xopt && argc - optind > 1) {
+        fprintf(stderr, "%s: -X only takes one directory name\n", prog);
+        exit(1);
+    }
+    if (opt.dirent && argc - optind > 1) {
+        fprintf(stderr, "%s: -D can only be used with one file\n", prog);
+        exit(1);
+    }
+
+    if (!opt.seq)
+        opt.seq = seq_lookup("nnsa");
+    assert(opt.seq != NULL);
+    printf("%s: using %s patterns\n", prog, opt.seq->desc);
+
+    if (opt.nohwrand)
+        disable_hwrand();
+    if (opt.nothreads)
+        disable_threads();
+
+    /* Scrub free space
+     */
+    if (Xopt) {
+        if (filetype(argv[optind]) != FILE_NOEXIST) {
+            fprintf(stderr, "%s: -X directory already exists\n", prog);
+            exit(1);
+        }
+        if (opt.dirent) {
+            fprintf(stderr, "%s: -D and -X cannot be used together\n", prog);
+            exit(1);
+        }
+        if (nopt) {
+            printf("%s: (dryrun) scrub free space in %s\n", prog, argv[optind]);
+        } else {
+            scrub_free(argv[optind], &opt);
+        }
+    /* Scrub multiple files/devices
+     */
+    } else if (argc - optind > 1) {
+        int i, errcount = 0;
+        for (i = optind; i < argc; i++)
+            errcount += scrub_object(argv[i], &opt, true, false);
+        if (errcount > 0) {
+            fprintf (stderr, "%s: no files were scrubbed\n", prog);
+            exit(1); 
+        }
+        for (i = optind; i < argc; i++) {
+            if (scrub_object(argv[i], &opt, false, nopt) > 0)
+                exit(1);
+        }
+    /* Scrub single file/device.
+     */
+    } else {
+        if (scrub_object(argv[optind], &opt, false, nopt) > 0)
+            exit(1);
+    }
+
+    if (custom_seq)
+        seq_destroy (custom_seq);
+    exit(0);
+}
+
+static int scrub_object(char *filename, const struct opt_struct *opt,
+                         bool noexec, bool dryrun)
+{
+    bool havesig = false;
+    int errcount = 0;
+
+    switch (filetype(filename)) {
+        case FILE_NOEXIST:
+            fprintf(stderr, "%s: %s does not exist\n", prog, filename);
+            errcount++;
+            break;
+        case FILE_OTHER:
+            fprintf(stderr, "%s: %s is wrong type of file\n", prog, filename);
+            errcount++;
+            break;
+        case FILE_BLOCK:
+        case FILE_CHAR:
+            if (opt->dirent) {
+                fprintf(stderr, "%s: cannot use -D with special file\n", prog);
+                errcount++;
+            } else if (opt->remove) {
+                fprintf(stderr, "%s: cannot use -r with special file\n", prog);
+                errcount++;
+            } else if (access(filename, R_OK|W_OK) < 0) {
+                fprintf(stderr, "%s: no rw access to %s\n", prog, filename);
+                errcount++;
+            } else if (checksig(filename, &havesig) < 0) {
+                fprintf(stderr, "%s: %s: %s\n", prog, filename,
+                        strerror(errno));
+                errcount++;
+            } else if (havesig && !opt->force) {
+                fprintf(stderr, "%s: %s already scrubbed? (-f to force)\n",
+                        prog, filename);
+                errcount++;
+            } else if (!noexec) {
+                if (dryrun) {
+                    printf("%s: (dryrun) scrub special file %s\n",
+                            prog, filename);
+                } else {
+                    scrub_disk(filename, opt);
+                }
+            }
+            break;
+        case FILE_LINK:
+            if (opt->nofollow) {
+                if (opt->remove && !noexec) {
+                    if (dryrun) {
+                        printf("%s: (dryrun) unlink %s\n", prog, filename);
+                    } else {
+                        printf("%s: unlinking %s\n", prog, filename);
+                        if (unlink(filename) != 0) {
+                            fprintf(stderr, "%s: unlink %s: %s\n", prog,
+                                    filename, strerror(errno));
+                            exit(1);
+                        }
+                    }
+                }
+                break;
+            }
+            /* FALL THRU */
+        case FILE_REGULAR:
+            if (access(filename, R_OK|W_OK) < 0) {
+                fprintf(stderr, "%s: no rw access to %s\n", prog, filename);
+                errcount++;
+            } else if (checksig(filename, &havesig) < 0) {
+                fprintf(stderr, "%s: %s: %s\n", prog, filename,
+                        strerror(errno));
+                errcount++;
+            } else if (havesig && !opt->force) {
+                fprintf(stderr, "%s: %s already scrubbed? (-f to force)\n",
+                        prog, filename);
+                errcount++;
+            } else if (opt->dirent && opt->dirent[0] != '/' && filename[0] == '/') {
+                fprintf(stderr, "%s: %s should be a full path like %s\n",
+                        prog, opt->dirent, filename);
+                errcount++;
+            } else if (!noexec) {
+                if (dryrun) {
+                    printf("%s: (dryrun) scrub reg file %s\n", prog, filename);
+                } else {
+                    scrub_file(filename, opt);
+                }
+#if __APPLE__
+                if (dryrun) {
+                    printf("%s: (dryrun) scrub res fork of %s\n",
+                           prog, filename);
+                } else {
+                    scrub_resfork(filename, opt);
+                }
+#endif
+                if (opt->dirent) {
+                    if (dryrun) {
+                        printf("%s: (dryrun) scrub dirent %s\n",
+                               prog, filename);
+                    } else {
+                        scrub_dirent(filename, opt);
+                    }
+                }
+                if (opt->remove) {
+                    char *rmfile = opt->dirent ? opt->dirent : filename;
+                    if (dryrun) {
+                        printf("%s: (dryrun) unlink %s\n", prog, rmfile);
+                    } else {
+                        printf("%s: unlinking %s\n", prog, rmfile);
+                        if (unlink(rmfile) != 0) {
+                            fprintf(stderr, "%s: unlink %s: %s\n", prog, rmfile,
+                                    strerror(errno));
+                            exit(1);
+                        }
+                    }
+                }
+            }
+            break;
+    }
+    return errcount;
+}
+
+static int progress_col (const sequence_t *seq)
+{
+    int i, max = 0, col = 50;
+    for (i = 0; i < seq->len; i++)
+        if (seq->pat[i].len > max)
+            max = seq->pat[i].len;
+    if (max > 3)
+        col -= (2 * (max - 3));
+    return col;
+}
+
 /* Scrub 'path', a file/device of size 'size'.
  * Fill using the pattern sequence specified by 'seq'.
  * Use 'bufsize' length for I/O buffers.
@@ -312,14 +427,15 @@ usage(void)
  */
 static bool
 scrub(char *path, off_t size, const sequence_t *seq, int bufsize, 
-      bool Sopt, bool sparse, bool enospc)
+      bool nosig, bool sparse, bool enospc)
 {
     unsigned char *buf;
     int i;
     prog_t p;
     char sizestr[80];
     bool isfull = false;
-    off_t written;
+    off_t written, checked;
+    int pcol = progress_col(seq);
 
     if (!(buf = alloc_buffer(bufsize))) {
         fprintf(stderr, "%s: out of memory\n", prog);
@@ -329,41 +445,73 @@ scrub(char *path, off_t size, const sequence_t *seq, int bufsize,
     size2str(sizestr, sizeof(sizestr), size);
     printf("%s: scrubbing %s %s\n", prog, path, sizestr);
 
-    initrand();
+    if (initrand() < 0) {
+        fprintf (stderr, "%s: initrand: %s\n", prog, strerror(errno));
+        exit(1);
+    }
     for (i = 0; i < seq->len; i++) {
         if (i > 0)
             enospc = false;
         switch (seq->pat[i].ptype) {
             case PAT_RANDOM:
                 printf("%s: %-8s", prog, "random");
-                progress_create(&p, 50);
-                churnrand();
+                progress_create(&p, pcol);
+                if (churnrand() < 0) {
+                    fprintf(stderr, "%s: churnrand: %s\n", prog,
+                             strerror(errno));
+                    exit(1);
+                }
                 written = fillfile(path, size, buf, bufsize, 
                                    (progress_t)progress_update, p, 
                                    (refill_t)genrand, sparse, enospc);
+                if (written == (off_t)-1) {
+                    fprintf(stderr, "%s: %s: %s\n", prog, path,
+                             strerror(errno));
+                    exit(1);
+                }
                 progress_destroy(p);
                 break;
             case PAT_NORMAL:
                 printf("%s: %-8s", prog, pat2str(seq->pat[i]));
-                progress_create(&p, 50);
+                progress_create(&p, pcol);
                 memset_pat(buf, seq->pat[i], bufsize);
                 written = fillfile(path, size, buf, bufsize, 
                                    (progress_t)progress_update, p, 
                                    NULL, sparse, enospc);
+                if (written == (off_t)-1) {
+                    fprintf(stderr, "%s: %s: %s\n", prog, path,
+                             strerror(errno));
+                    exit(1);
+                }
                 progress_destroy(p);
                 break;
             case PAT_VERIFY:
                 printf("%s: %-8s", prog, pat2str(seq->pat[i]));
-                progress_create(&p, 50);
+                progress_create(&p, pcol);
                 memset_pat(buf, seq->pat[i], bufsize);
                 written = fillfile(path, size, buf, bufsize, 
                                    (progress_t)progress_update, p, 
                                    NULL, sparse, enospc);
+                if (written == (off_t)-1) {
+                    fprintf(stderr, "%s: %s: %s\n", prog, path,
+                             strerror(errno));
+                    exit(1);
+                }
                 progress_destroy(p);
                 printf("%s: %-8s", prog, "verify");
-                progress_create(&p, 50);
-                checkfile(path, written, buf, bufsize, 
-                          (progress_t)progress_update, p, sparse);
+                progress_create(&p, pcol);
+                checked = checkfile(path, written, buf, bufsize, 
+                                    (progress_t)progress_update, p, sparse);
+                if (checked == (off_t)-1) {
+                    fprintf(stderr, "%s: %s: %s\n", prog, path,
+                             strerror(errno));
+                    exit(1);
+                }
+                if (checked < written) {
+                    fprintf(stderr, "%s: %s: verification error\n",
+                             prog, path);
+                    exit(1);
+                }
                 progress_destroy(p);
                 break;
         }
@@ -373,14 +521,18 @@ scrub(char *path, off_t size, const sequence_t *seq, int bufsize,
             isfull = true; 
             size = written;
             if (size == 0) {
-                fprintf(stderr, "%s: file system is full (0 bytes written)\n",
-                        prog);
+                printf("%s: file system is full (0 bytes written)\n", prog);
                 break;
             }
         }
     }
-    if (!Sopt)
-        writesig(path);
+    if (!nosig && written > 0) {
+        if (writesig(path) < 0) {
+            fprintf(stderr, "%s: writing signature to %s: %s\n", prog,
+                    path, strerror (errno));
+            exit (1);
+        }
+    }
 
     free(buf);
     return isfull;
@@ -413,16 +565,16 @@ set_rlimit_fsize(off_t val)
 }
 
 /* Scrub free space (-X) by creating a directory, then filling it
- * with 'size' length files (use RLIMIT_FSIZE if no -s).
+ * with opt->devsize length files (use RLIMIT_FSIZE if no opt->devsize).
  */
 static void
-scrub_free(char *dirpath, off_t size, const sequence_t *seq, 
-           int bufsize, bool Sopt)
+scrub_free(char *dirpath, const struct opt_struct *opt)
 {
     char path[MAXPATHLEN];
     int fileno = 0;
     struct stat sb;
     bool isfull;
+    off_t size = opt->devsize;
 
     if (mkdir(dirpath, 0755) < 0) {
         fprintf(stderr, "%s: mkdir %s: %s\n", prog, path, strerror(errno));
@@ -442,26 +594,28 @@ scrub_free(char *dirpath, off_t size, const sequence_t *seq,
     size = blkalign(size, sb.st_blksize, DOWN);
     do {
         snprintf(path, sizeof(path), "%s/scrub.%.3d", dirpath, fileno++);
-        isfull = scrub(path, size, seq, bufsize, Sopt, false, true);
+        isfull = scrub(path, size, opt->seq, opt->blocksize, opt->nosig,
+                       false, true);
     } while (!isfull);
     while (--fileno >= 0) {
         snprintf(path, sizeof(path), "%s/scrub.%.3d", dirpath, fileno);
         if (unlink(path) < 0)
             fprintf(stderr, "%s: unlink %s: %s\n", prog, path, strerror(errno));
         else
-            fprintf(stderr, "%s: unlinked %s\n", prog, path);
+            printf("%s: unlinked %s\n", prog, path);
     }
     if (rmdir(dirpath) < 0)
         fprintf(stderr, "%s: rmdir %s: %s\n", prog, dirpath, strerror(errno));
     else
-        fprintf(stderr, "%s: removed %s\n", prog, dirpath);
+        printf("%s: removed %s\n", prog, dirpath);
 }
 
 /* Scrub name component of a directory entry through succesive renames.
  */
 static void
-scrub_dirent(char *path, char *newpath)
+scrub_dirent(char *path, const struct opt_struct *opt)
 {
+    char *newpath = opt->dirent;
     const sequence_t *seq = seq_lookup("dirent");
     prog_t p;
     int i;
@@ -476,8 +630,11 @@ scrub_dirent(char *path, char *newpath)
         assert(seq->pat[i].ptype == PAT_NORMAL);
         assert(seq->pat[i].len == 1);
         printf("%s: %-8s", prog, pat2str(seq->pat[i]));
-        progress_create(&p, 50);
-        filldentry(path, seq->pat[i].pat[0]); /* path: in/out */
+        progress_create(&p, progress_col(seq));
+        if (filldentry(path, seq->pat[i].pat[0]) < 0) {/* path: in/out */
+            fprintf(stderr, "%s: filldentry: %s\n", prog, strerror(errno));
+            exit(1);
+        } 
         progress_update(p, 1.0);
         progress_destroy(p);
     }
@@ -490,11 +647,11 @@ scrub_dirent(char *path, char *newpath)
 /* Scrub a regular file.
  */
 static void 
-scrub_file(char *path, off_t size, const sequence_t *seq, 
-           int bufsize, bool Sopt, bool sparse)
+scrub_file(char *path, const struct opt_struct *opt)
 {
     struct stat sb;
     filetype_t ftype = filetype(path);
+    off_t size = opt->devsize;
 
     assert(ftype == FILE_REGULAR || ftype == FILE_LINK);
 
@@ -516,14 +673,14 @@ scrub_file(char *path, off_t size, const sequence_t *seq,
                     prog, path, (int)(size - sb.st_size)); 
         }
     }
-    scrub(path, size, seq, bufsize, Sopt, sparse, false);
+    scrub(path, size, opt->seq, opt->blocksize, opt->nosig, opt->sparse, false);
 }
 
 /* Scrub apple resource fork component of file.
  */
 #if __APPLE__
 static void
-scrub_resfork(char *path, const sequence_t *seq, int bufsize)
+scrub_resfork(char *path, const struct opt_struct *opt)
 {
     struct stat rsb;
     char rpath[MAXPATHLEN];
@@ -544,28 +701,29 @@ scrub_resfork(char *path, const sequence_t *seq, int bufsize)
         printf("%s: padding %s with %d bytes to fill last fs block\n", 
                         prog, rpath, (int)(rsize - rsb.st_size)); 
     }
-    scrub(rpath, rsize, seq, bufsize, false, false, false);
+    scrub(rpath, rsize, opt->seq, opt->blocksize, false, false, false);
 }
 #endif
 
 /* Scrub a special file corresponding to a disk.
  */
 static void
-scrub_disk(char *path, off_t size, const sequence_t *seq, int bufsize, 
-           bool Sopt, bool sparse)
+scrub_disk(char *path, const struct opt_struct *opt)
 {
     filetype_t ftype = filetype(path);
+    off_t devsize = opt->devsize;
 
     assert(ftype == FILE_BLOCK || ftype == FILE_CHAR);
-    if (size == 0) {
-        size = getsize(path);
-        if (size == 0) {
+    if (devsize == 0) {
+        if (getsize(path, &devsize) < 0) {
+            fprintf(stderr, "%s: %s: %s\n", prog, path, strerror(errno));
             fprintf(stderr, "%s: could not determine size, use -s\n", prog);
             exit(1);
         }
         printf("%s: please verify that device size below is correct!\n", prog);
     }
-    scrub(path, size, seq, bufsize, Sopt, sparse, false);
+    scrub(path, devsize, opt->seq, opt->blocksize, opt->nosig, opt->sparse,
+          false);
 }
 
 /*
